This document explains how to profile the memory usage of go-dqlite using a synthetic load.

The synthetic load and profiling is done via a script that utilizes the `dqlite-demo` program under go-dqlite. The script first creates a 3 member dqlite cluster and starts inserting key value pairs with a curl call, size of the values being set by `--value-size` and creates N number of keys determined by `--iterations`.

To get started follow the guide on [go-dqlite](https://github.com/canonical/go-dqlite#demo) for demo.
To replace the binary from the deb packages do
```bash
sudo rm -rf /usr/bin/dqlite-demo
sudo ln -s "$HOME/go/bin/dqlite-demo" "/usr/bin/dqlite-demo"
```
Install the required software with
```bash
sudo apt install valgrind
go install github.com/google/pprof@latest
sudo apt install debian-goodies
sudo apt install smem
```
To run the `run-memory-profiling.sh` script 
```bash
sudo ./load.sh --iterations 1000
```
and for disk mode
```
sudo ./load.sh --iterations 1000 --disk-mode
```

The script also has different options to profile stacks and the whole memory pages
```
sudo ./load.sh --iterations 1000 --stacks
sudo ./load.sh --iterations 1000 --pages
```

There will be different files outputted under folders like `dqlite-demo-1000`.
The files named `dqlite-demo-leader`, `dqlite-demo-voter-1`, `dqlite-demo-voter-2` are Massif output files and can be visualized using [Massif Visualizer](https://apps.kde.org/massif-visualizer/).
The CSV files are captured outputs of system reported memory by smem, and can be imported into a Sheet program to create charts/graphs.

Memory profiles are generated by using valgrind with massif. There are a number of options that can be used with massif that will let valgrind
- `Default` Profile the memory heap, only by replacing malloc and similar calls
- `--stacks=yes` Profile the memory heap, by replacing malloc and similar calls and also profiling the stack
- `--pages-as-heap=yes` Profile the whole memory used, via observing the memory pages that goes down to lower level mmap calls

The above mentioned first 2 options will not include memory handled by the go runtime, which will only profile the linked C libraries of dqlite. Although the 3rd option will observe the whole memory, valgrind can not analyze go code. 

Valgrind also requires proper debug-symbols in place to accurately create a callstack in profiling. To find the required debug symbols packages, there is a tool called `find-dbgsym-packages`. Which can be installed via `apt install debian-goodies`. By running this command the tool will list available packages that can be installed via apt and will also libraries with missing dbg or dbgsym packages. For example dbgsym packages for `go-dqlite` is not available through apt even through the `https://launchpad.net/~dqlite/+archive/ubuntu/dev/+packages` ppa. So the ddeb packages for 
libdqlite0-dbgsym
libraft3-dbgsym
libsqlite3-0-dbgsym
Needs to be downloaded manually through launchpad. After downloading the ddeb packages, a simple `dpkg -i libdqlite0-dbgsym.ddeb` command would suffice. Be aware that matching the version numbers down to the patch level is necessary.

To verify the profiling data from valgrind, the system reported memory usage is also measured using smem. To get started with smem, install it with `apt install smem`. Smem reports 3 different columns for memory usage, listed below:
- USS - Unique Set Size. This is the amount of unshared memory unique to that process (think of it as U for unique memory). It does not include shared memory. Thus this will under-report the amount of memory a process uses, but it is helpful when you want to ignore shared memory.
- PSS - Proportional Set Size. This is what you want. It adds together the unique memory (USS), along with a proportion of its shared memory divided by the number of processes sharing that memory. Thus it will give you an accurate representation of how much actual physical memory is being used per process - with shared memory truly represented as shared. Think of the P being for physical memory.
- RSS - Resident Set Size. This is the amount of shared memory plus unshared memory used by each process. If any processes share memory, this will over-report the amount of memory actually used, because the same shared memory will be counted more than once - appearing again in each other process that shares the same memory.

To profile the Go code of `go-dqlite`, the `pprof` tool needs to be utilized. To use pprof
Add the pprof library to the program via `import _ "net/http/pprof"`
Create a http listener if there isn’t one already, e.g. `log.Println(http.ListenAndServe("localhost:6060", nil))` (There is one for dqlite-demo already listening at 8001 on the leader)
Open the pprof web ui to visualize heap profile, e.g. `pprof -http=0.0.0.0:15345 /usr/bin/dqlite-demo http://127.0.0.1:8001/debug/pprof/heap`
Pprof will create a profile with samples that includes flat and cumulative memory usage.
